#include "create_cic_fsc.h"

Create_CIC_FSC::Create_CIC_FSC(int argc, char **argv)
{
    unsigned char hexvalue = 0;

    if (argc != 4 && argc != 6)
    {
        usage();
        return;
    }

    if (!validateFile(argv[1]))
    {
        cout << "Could not open " << argv[1] << endl;
        return;
    }

    if (argc == 4)
    {
        regionhigh = 0;
        if (!validateHex(argv[2], regionlow))
        {
            cout << argv[2] << " is not a valid hex number" << endl;
            return;
        }

        yearhigh = 0;
        if (!validateHex(argv[3], yearlow))
        {
            cout << argv[3] << " is not a valid hex number" << endl;
            return;
        }
    }
    else //argc == 6
    {
        if (!validateHex(argv[2], regionhigh))
        {
            cout << argv[2] << " is not a valid hex number" << endl;
            return;
        }
        if (!validateHex(argv[3], regionlow))
        {
            cout << argv[3] << " is not a valid hex number" << endl;
            return;
        }

        if (!validateHex(argv[4], yearhigh))
        {
            cout << argv[4] << " is not a valid hex number" << endl;
            return;
        }
        if (!validateHex(argv[5], yearlow))
        {
            cout << argv[5] << " is not a valid hex number" << endl;
            return;
        }
    }

    int rsafile1blocn = find1BHeader();
    if (rsafile1blocn == -1)
    {
        cout << "No CIC FSC (1B) found in " << argv[1] << endl;
        return;
    }

    rsa1bfile = (rsa1bstruct *)&rsafile[rsafile1blocn];         //1B is a fixed size struct
    cout << "CIC FSC (1B) found for VIN: " << rsa1bfile->vin << " with timestamp: " << rsa1bfile->timestamp << endl;

    if (rsa1bfile->separator != 0x00)
    {
        cout << "RSA Key is not 128 bytes\nKey corrupt" << endl;
        return;
    }


    encrypted = BN_new();
    unsigned char des_cblock_key[MD5_DIGEST_LENGTH / 2] = {0};
    unsigned char enc_msg[24] = {0};

    calc_RSA();
    calc_MD5(des_cblock_key);
    calc_DES(des_cblock_key, enc_msg);
    calc_Code(enc_msg);
}

Create_CIC_FSC::~Create_CIC_FSC()
{

}

int Create_CIC_FSC::usage(void)
{
    cout << "\nUsage:  ./create_cic_fsc CICRSAFile.bin TYPE [TYPE] YEAR [YEAR]" << endl;
    cout << "\n\tTYPE 1 or 2 bytes representing the map type" << endl;
    cout << "\t\t   If TYPE is two bytes, YEAR must be two bytes" << endl;
    cout << "\tYEAR 1 or 2 bytes representing the map year" << endl;
    cout << "\t\t   If YEAR is two bytes, TYPE must be two bytes" << endl;
    cout << "\nExample ./create_cic_fsc dump03.bin 0x2D 0x4D" << endl;
    cout << "\nMap TYPE and YEAR can be found in the Lookup.xml file (or online)" << endl;

    return 0;
}

bool Create_CIC_FSC::validateFile(string inFilename)
{
    fstream inFile(inFilename, ios::in | ios::binary);
    if(inFile.is_open())
    {
        inFile.seekg(0, ios::end);
        rsafile_size = inFile.tellg();
        inFile.seekg(0, ios::beg);

        rsafile = new char[rsafile_size];
        inFile.read(rsafile, rsafile_size);

        return true;
    }
    return false;
}

bool Create_CIC_FSC::validateHex(string inHex, unsigned char &yearregion)
{
    transform(inHex.begin(), inHex.end(), inHex.begin(), ::toupper);

    if (!(inHex.compare(0, 2, "0X")))
    {
        inHex.erase(0, 2);
    }

    if (inHex.length() == 2)
    {
        if (::all_of(inHex.begin(), inHex.end(), ::isxdigit))
        {
            yearregion = ::stoi(inHex, 0, 16);
            return true;
        }
    }
    return false;
}

int Create_CIC_FSC::find1BHeader(void)
{
    int i = 0;
    for (i = 0; i < (rsafile_size - 4); i++)
    {
        bool flag = memcmp(&rsafile[i], &header1b, sizeof(header1b));

        if (!flag)
        {
            return i;
        }
    }
    return -1;
}

void Create_CIC_FSC::calc_RSA(void)
{
    BIGNUM *message = BN_new();
    BIGNUM *rsa_e = BN_new();
    BIGNUM *modulo = BN_new();
    bignum_ctx *ctx = BN_CTX_new();

    BN_lebin2bn(rsa1bfile->rsakey01, RSA_LENGTH, message);
    BN_hex2bn(&rsa_e, RSA_E_CIC);
    BN_hex2bn(&modulo, MODULO_CIC);

    BN_mod_exp(encrypted, message, rsa_e, modulo, ctx);
}

void Create_CIC_FSC::calc_MD5(unsigned char *des_cblock_key)
{
    unsigned char digest[MD5_DIGEST_LENGTH];
    unsigned int digest_len;
    unsigned char encrypted_array[RSA_LENGTH] = {0};
    unsigned char memcpy_position = 0;
    unsigned char md5_two[MD5_TWO_LENGTH] = {0};
    unsigned char digest2[MD5_DIGEST_LENGTH];
    unsigned int digest2_len;
    unsigned char md5_three[MD5_THREE_LENGTH] = {0};
    unsigned int digest3_len;
    unsigned char digest3[MD5_DIGEST_LENGTH];
    unsigned char digest_old[MD5_DIGEST_LENGTH];
    unsigned char digest_new[MD5_DIGEST_LENGTH];
    unsigned int digest_newlen;

    /////////////////////////////// First MD5 Routine ///////////////////////////////

    BN_bn2lebinpad(encrypted, encrypted_array, BN_num_bytes(encrypted));

    EVP_MD_CTX *mdctx = EVP_MD_CTX_new();
    const EVP_MD *md = EVP_md5();

    EVP_DigestInit_ex(mdctx, md, NULL);
    EVP_DigestUpdate(mdctx, encrypted_array, (sizeof encrypted_array / sizeof *encrypted_array));
    EVP_DigestFinal_ex(mdctx, digest, &digest_len);

    /////////////////////////////// Second MD5 Routine ///////////////////////////////

    memcpy(md5_two, rsa1bfile->vin, VIN_LENGTH);
    memcpy_position = VIN_LENGTH;
    memcpy(md5_two + VIN_LENGTH, digest, 8);
    memcpy_position += 8;
    md5_two[memcpy_position] = regionhigh;
    memcpy_position++;
    md5_two[memcpy_position] = regionlow;
    memcpy_position++;
    md5_two[memcpy_position] = yearhigh;
    memcpy_position++;
    md5_two[memcpy_position] = yearlow;

    EVP_DigestInit_ex(mdctx, md, NULL);
    EVP_DigestUpdate(mdctx, md5_two, (sizeof md5_two / sizeof *md5_two));
    EVP_DigestFinal_ex(mdctx, digest2, &digest2_len);

    /////////////////////////////// Third MD5 Routine ///////////////////////////////

    memcpy(md5_three, digest, 8);
    memcpy_position = 8;
    memcpy(md5_three + memcpy_position, digest2, 8);

    EVP_DigestInit_ex(mdctx, md, NULL);
    EVP_DigestUpdate(mdctx, md5_three, (sizeof md5_three / sizeof *md5_three));
    EVP_DigestFinal_ex(mdctx, digest3, &digest3_len);

    memcpy(digest_old, digest3, (sizeof digest3 / sizeof *digest3));

    for (int i = 1; i < 8; i++)
    {
        EVP_DigestInit_ex(mdctx, md, NULL);
        EVP_DigestUpdate(mdctx, digest_old, (sizeof digest_old / sizeof *digest_old));
        EVP_DigestFinal_ex(mdctx, digest_new, &digest_newlen);
        memcpy(digest_old, digest_new, (sizeof digest_new / sizeof *digest_new));
    }
    EVP_MD_CTX_free(mdctx);

    for (int i = 0 ; i < 8; i++)
    {
        des_cblock_key[i] = digest_old[i] ^ digest_old[i + 8];
    }
}

void Create_CIC_FSC::calc_DES(unsigned char *des_cblock_key, unsigned char *enc_msg)
{
    unsigned char des_key[DES_LENGTH] = {0};
    unsigned char msg[] = {0x01, regionhigh, regionlow, yearhigh, yearlow, 0x01, 0x00, 0x00};
    int outl = 0, ciphertext_total = 0;

    memcpy(des_key, des_cblock_key, (sizeof des_key / sizeof *des_key));

    /////////////////////////////// OpenSSL considers DES legacy ///////////////////////////////
    //////////////////// The algorithm does not work if Provider is not set ////////////////////

    OSSL_PROVIDER *legacy;
    legacy = OSSL_PROVIDER_load(NULL, "legacy");
    if (legacy == NULL)
    {
        cout << "OpenSSL could not load DES legacy algorithms" << endl;
    }

    const EVP_CIPHER *evp_ecb = EVP_des_ecb();
    EVP_CIPHER_CTX *ctx = EVP_CIPHER_CTX_new();
    EVP_EncryptInit(ctx, evp_ecb, des_key, NULL);
    EVP_EncryptUpdate(ctx, enc_msg, &outl, msg, sizeof(msg));
    ciphertext_total += outl;
    EVP_EncryptFinal(ctx, enc_msg + ciphertext_total, &outl);
}

void Create_CIC_FSC::calc_Code(unsigned char *enc_msg)
{
    unsigned char finalhash[FINALHASH_LENGTH] = {0};
    const unsigned char lookuparray[] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '2', '3', '4', '5', '6', '7'};
    unsigned int shiftright = 0, carry = 8, finalhashposition = 0;
    unsigned int currentvalue = enc_msg[finalhashposition];
    unsigned int codevalue = 0, shiftedvalue = 0;
    unsigned char memcpy_position = 0;
    string bmwcode;

    memcpy(finalhash, enc_msg, 8);
    memcpy_position = 8;
    finalhash[memcpy_position] = regionhigh;
    memcpy_position++;
    finalhash[memcpy_position] = regionlow;
    memcpy_position++;
    finalhash[memcpy_position] = yearhigh;
    memcpy_position++;
    finalhash[memcpy_position] = yearlow;

    for (int i = 0; i < 20; i++)
    {
        shiftright = carry - 5;
        shiftedvalue  = currentvalue >> shiftright;
        codevalue = lookuparray[(shiftedvalue & 0x1F)];
        bmwcode += (unsigned char)codevalue;

        if (shiftright > 4)
        {
            carry = shiftright;
        }
        else
        {
            finalhashposition++;
            currentvalue = currentvalue << 8;
            currentvalue = currentvalue | finalhash[finalhashposition];
            carry += 3;
        }
    }

    cout << "BMW CIC Navigation Code: " << bmwcode << endl;
}
